package it.inaf.oats.ia2.tapschemamodel;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

/**
 * Object that keeps track of changes in the value that it wraps.
 */
class EditableValue {

    String originalValue;
    String newValue;

    EditableValue(String value) {
        setValue(value);
    }

    String getValue() {
        return newValue;
    }

    final void setValue(String value) {
        originalValue = value;
        newValue = value;
    }

    void updateValue(String value) {
        newValue = value;
    }

    boolean isChanged() {
        if (originalValue == null) {
            return newValue != null && !newValue.equals("");
        }
        return !originalValue.equals(newValue);
    }
}

/**
 *
 * @author Sonia Zorba <zorba at oats.inaf.it>
 */
public abstract class DbEntity {

    private Status status;

    private final HashMap<String, EditableValue> values = new HashMap<String, EditableValue>();

    public DbEntity(String... supportedValues) {
        for (String key : supportedValues) {
            setValue(key, null);
        }
    }

    public final void setValue(String key, String value) {
        EditableValue editableValue = values.get(key);
        if (editableValue == null) {
            editableValue = new EditableValue(value);
            values.put(key, editableValue);
        } else {
            editableValue.setValue(value);
        }
    }

    public boolean updateValue(String key, String value) {
        EditableValue editableValue = values.get(key);
        editableValue.updateValue(value);
        return editableValue.isChanged();
    }

    public String getValue(String key) {
        EditableValue editableValue = values.get(key);
        if (editableValue != null) {
            return editableValue.getValue();
        }
        return null;
    }

    public boolean isChanged(String key) {
        EditableValue editableValue = values.get(key);
        if (editableValue != null) {
            return editableValue.isChanged();
        }
        return false;
    }

    public Collection<String> getSupportedValues() {
        return values.keySet();
    }

    public Status getStatus() {
        return status;
    }

    public void setStatus(Status status) {
        this.status = status;
    }

    public boolean hasChanges() {
        for (EditableValue value : values.values()) {
            if (value.isChanged()) {
                return true;
            }
        }
        return false;
    }

    public Map<String, Object> getChanges() {
        Map<String, Object> changes = new HashMap<String, Object>();
        for (String key : values.keySet()) {
            EditableValue editableValue = values.get(key);
            if (editableValue != null) {
                changes.put(key, editableValue.getValue());
            }
        }
        return changes;
    }
    
    public void setUpdated() {
        if (getStatus() == Status.SELECTED_TO_SAVE) {
            setStatus(Status.SELECTED_SAVED);
        }
        for (EditableValue value : values.values()) {
            value.setValue(value.getValue());
        }
    }
    
    public void setOriginalValues() {
        for (EditableValue value : values.values()) {
            value.newValue = value.originalValue;
        }
    }

    public abstract String getName();

    public abstract void insert(Connection connection) throws SQLException;

    public abstract void update(Connection connection) throws SQLException;
}
