package it.inaf.oats.ia2.tapschemamodel;

import java.io.Serializable;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 *
 * @author pietro
 */
public class TapSchema implements Serializable, IDbEntitiesContainer {

    private final String name;
    private boolean toCreate;
    //private final ArrayList<String> allSchemas;

    // current schemas (saved, added or to remove)
    private final Map<String, Schema> schemasMap;

    private String selectedSchema;

    // For EXISTING TAP Schemas
    public TapSchema(Connection connection, String tapSchemaName) throws SQLException {
        this(connection, tapSchemaName, false, null);
        toCreate = false;
    }

    // For NEW TAP Schemas
    public TapSchema(Connection connection, String tapSchemaName, ArrayList<String> selectedSchemas) throws SQLException {
        this(connection, tapSchemaName, true, selectedSchemas);
    }

    private TapSchema(final Connection connection, String tapSchemaName, boolean toCreate, ArrayList<String> selectedSchemas) throws SQLException {
        this.name = tapSchemaName;
        this.toCreate = toCreate;

        schemasMap = new HashMap<String, Schema>();
        for (String schemaName : TapSchemaDL.getAllSchemasNames(connection)) {
            schemasMap.put(schemaName, null);
        }

        if (toCreate) {
            loadEntities(connection, selectedSchemas);
        } else {
            TapSchemaDL.loadSavedSchemas(connection, (this));
        }

        if (!schemasMap.isEmpty()) {
            for (String key : schemasMap.keySet()) {
                Schema schema = schemasMap.get(key);
                if (schema != null) {
                    select(schema.getName());
                    break;
                }
            }
        }
    }

    @Override
    public final void loadEntities(Connection connection, List<String> selectedSchemas) throws SQLException {
        TapSchemaDL.loadSelectedSchemas(connection, this, selectedSchemas);
        select(selectedSchemas.get(selectedSchemas.size() - 1));
    }

    public boolean toCreate() {
        return toCreate;
    }

    public Schema getSchema(String name) {
        return schemasMap.get(name);
    }

    @Override
    public void addEntity(String name, DbEntity schema) {
        schemasMap.put(name, (Schema) schema);
    }

    public Collection<String> getSchemasNames() {
        return schemasMap.keySet();
    }

    @Override
    public ArrayList<DbEntity> getAllEntities() {
        ArrayList<DbEntity> list = new ArrayList<DbEntity>();
        for (DbEntity schema : schemasMap.values()) {
            if (schema != null) {
                list.add(schema);
            }
        }
        return list;
    }

    public String getName() {
        return name;
    }

    @Override
    public ArrayList<String> getAddables() {
        ArrayList<String> addableSchemas = new ArrayList<String>();

        for (String schemaName : schemasMap.keySet()) {
            if (schemasMap.get(schemaName) == null && !schemaName.equals(this.name)) {
                addableSchemas.add(schemaName);
            }
        }

        return addableSchemas;
    }

    @Override
    public void removeEntity(String schemaName) {
        Schema schema = getSchema(schemaName);
        if (schema.getStatus() == Status.SELECTED_TO_SAVE) {
            schemasMap.put(schemaName, null);
            ArrayList<DbEntity> entities = getAllEntities();
            if (!entities.isEmpty()) {
                select(entities.get(0).getName());
            }
        }
        if (schema.getStatus() == Status.SELECTED_SAVED) {
            schema.setStatus(Status.TO_REMOVE);
        }
    }

    @Override
    public DbEntity getSelected() {
        return schemasMap.get(selectedSchema);
    }

    @Override
    public void select(String name) {
        selectedSchema = name;

        // select first table
        ArrayList<DbEntity> tables = getSelectedSchema().getAllEntities();
        if (!tables.isEmpty()) {
            getSelectedSchema().select(tables.get(0).getName());
        }
    }

    public Schema getSelectedSchema() {
        if (getSelected() == null) {
            return null;
        }
        return (Schema) getSelected();
    }

    public Table getSelectedTable() {
        if (getSelected() != null) {
            Object table = ((Schema) getSelected()).getSelected();
            if (table != null) {
                return (Table) table;
            }
        }
        return null;
    }

    public Column getSelectedColumn() {
        Table table = getSelectedTable();
        if (table != null) {
            Object column = table.getSelected();
            if (column != null) {
                return (Column) column;
            }
        }
        return null;
    }

    @Override
    public void update() {
        for (DbEntity schemaEntity : getAllEntities()) {
            Schema schema = (Schema) schemaEntity;
            if (schema.getStatus() == Status.TO_REMOVE) {
                schemasMap.put(schema.getName(), null);
            } else {
                schema.update();
            }
        }
    }
}
