package it.inaf.oats.ia2.tapschemamodel;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.HashSet;

/**
 *
 * @author Sonia Zorba <zorba at oats.inaf.it>
 */
public class UpdateManager {

    private final Credentials credentials;
    private final TapSchema tapSchema;

    private final ArrayList<Integer> keysToRemove;

    private final ArrayList<Schema> schemasToAdd;
    private final ArrayList<Table> tablesToAdd;
    private final ArrayList<Column> columnsToAdd;

    private final ArrayList<String> schemasToRemove;
    private final ArrayList<String> tablesToRemove;
    private final ArrayList<String> columnsToRemove;

    private final ArrayList<Schema> schemasToUpdate;
    private final ArrayList<Table> tablesToUpdate;
    private final ArrayList<Column> columnsToUpdate;

    public UpdateManager(Credentials credentials, TapSchema tapSchema) {
        this.credentials = credentials;
        this.tapSchema = tapSchema;

        keysToRemove = new ArrayList<Integer>();

        schemasToAdd = new ArrayList<Schema>();
        tablesToAdd = new ArrayList<Table>();
        columnsToAdd = new ArrayList<Column>();

        schemasToRemove = new ArrayList<String>();
        tablesToRemove = new ArrayList<String>();
        columnsToRemove = new ArrayList<String>();

        schemasToUpdate = new ArrayList<Schema>();
        tablesToUpdate = new ArrayList<Table>();
        columnsToUpdate = new ArrayList<Column>();
    }

    private void createTapSchema(Connection connection) throws SQLException {
        Statement statement = null;

        try {
            statement = connection.createStatement();

            String tapschemaName = tapSchema.getName();

            //creates database
            String queryString = "CREATE DATABASE IF NOT EXISTS " + tapschemaName + ";";
            statement.executeUpdate(queryString);

            //creates schemas table
            queryString = "CREATE TABLE IF NOT EXISTS " + tapschemaName + ".`schemas` ("
                    + "schema_name   varchar(64), "
                    + "utype         varchar(512) NULL, "
                    + "description   varchar(512) NULL, "
                    + "schemaID      bigint, "
                    + "PRIMARY KEY (schema_name));";
            statement.executeUpdate(queryString);

            //creates tables table
            queryString = "CREATE TABLE IF NOT EXISTS " + tapschemaName + ".tables ("
                    + "schema_name   varchar(64), "
                    + "table_name    varchar(128), "
                    + "table_type    varchar(8), "
                    + "utype         varchar(512) NULL, "
                    + "description   varchar(512) NULL, "
                    + "tableID       bigint, "
                    + "PRIMARY KEY (table_name), "
                    + "FOREIGN KEY (schema_name) REFERENCES " + tapschemaName + ".`schemas` (schema_name));";
            statement.executeUpdate(queryString);

            //creates columns table
            queryString = "CREATE TABLE IF NOT EXISTS " + tapschemaName + ".columns ("
                    + "table_name    varchar(128), "
                    + "column_name   varchar(64), "
                    + "utype         varchar(512) NULL, "
                    + "ucd           varchar(64)  NULL, "
                    + "unit          varchar(64)  NULL, "
                    + "description   varchar(512) NULL, "
                    + "datatype      varchar(64)  NOT NULL, "
                    + "size          integer      NULL, "
                    + "principal     integer      NOT NULL, "
                    + "indexed       integer      NOT NULL, "
                    + "std           integer      NOT NULL, "
                    + "columnID      bigint, "
                    + "PRIMARY KEY (table_name, column_name), "
                    + "FOREIGN KEY (table_name) REFERENCES " + tapschemaName + ".tables (table_name));";
            statement.executeUpdate(queryString);

            //creates keys table
            queryString = "CREATE TABLE IF NOT EXISTS " + tapschemaName + ".keys ("
                    + "key_id        varchar(64), "
                    + "from_table    varchar(128) NOT NULL, "
                    + "target_table  varchar(128) NOT NULL, "
                    + "utype         varchar(512) NULL, "
                    + "description   varchar(512) NULL, "
                    + "keyID         bigint, "
                    + "PRIMARY KEY (key_id), "
                    + "FOREIGN KEY (from_table) REFERENCES " + tapschemaName + ".tables (table_name), "
                    + "FOREIGN KEY (target_table) REFERENCES " + tapschemaName + ".tables (table_name));";
            statement.executeUpdate(queryString);

            //creates key_columns table
            queryString = "CREATE TABLE IF NOT EXISTS " + tapschemaName + ".key_columns ("
                    + "key_id          varchar(64), "
                    + "from_column     varchar(64)   NOT NULL, "
                    + "target_column   varchar(64) NOT NULL, "
                    + "key_columnID    bigint, "
                    + "FOREIGN KEY (key_id) REFERENCES " + tapschemaName + ".keys (key_id));";
            statement.executeUpdate(queryString);
        } catch (SQLException e) {
            throw e;
        } finally {
            if (statement != null) {
                statement.close();
            }
        }
    }

    public void update() {

        // Separate changes in specific lists
        for (DbEntity schemaDbEntity : tapSchema.getAllEntities()) {
            Schema schema = (Schema) schemaDbEntity;

            if (schema.getStatus() != Status.NOT_SELECTED) {

                switch (schema.getStatus()) {
                    case TO_REMOVE:
                        schemasToRemove.add(schema.getName());
                        break;
                    case SELECTED_TO_SAVE:
                        schemasToAdd.add(schema);
                        break;
                    case SELECTED_SAVED:
                        if (schema.hasChanges()) {
                            schemasToUpdate.add(schema);
                        }
                        break;
                }

                for (DbEntity tableDbEntity : schema.getAllEntities()) {
                    Table table = (Table) tableDbEntity;

                    if (table.getStatus() != Status.NOT_SELECTED) {

                        String tableName = table.getFullName();

                        switch (table.getStatus()) {
                            case TO_REMOVE:
                                tablesToRemove.add(tableName);
                                break;
                            case SELECTED_TO_SAVE:
                                if (schema.getStatus() == Status.SELECTED_SAVED || schema.getStatus() == Status.SELECTED_TO_SAVE) {
                                    tablesToAdd.add(table);
                                }
                                break;
                            case SELECTED_SAVED:
                                if (schema.getStatus() == Status.TO_REMOVE) {
                                    tablesToRemove.add(tableName);
                                } else if (table.hasChanges()) {
                                    tablesToUpdate.add(table);
                                }
                                break;
                        }

                        for (DbEntity columnDbEntity : table.getAllEntities()) {
                            Column column = (Column) columnDbEntity;

                            if (column.getStatus() != Status.NOT_SELECTED) {

                                String columnName = column.getName();

                                switch (column.getStatus()) {
                                    case TO_REMOVE:
                                        columnsToRemove.add(columnName);
                                        break;
                                    case SELECTED_TO_SAVE:
                                        if ((schema.getStatus() == Status.SELECTED_SAVED || schema.getStatus() == Status.SELECTED_TO_SAVE)
                                                && (table.getStatus() == Status.SELECTED_SAVED || table.getStatus() == Status.SELECTED_TO_SAVE)) {
                                            columnsToAdd.add(column);
                                        }
                                        break;
                                    case SELECTED_SAVED:
                                        if (schema.getStatus() == Status.TO_REMOVE || table.getStatus() == Status.TO_REMOVE) {
                                            columnsToRemove.add(columnName);
                                        } else if (column.hasChanges()) {
                                            columnsToUpdate.add(column);
                                        }
                                        break;
                                }
                            }
                        }
                    }
                }
            }
        }

        Connection connection = null;
        PreparedStatement statement = null;
        ResultSet resultSet = null;

        try {

            connection = credentials.getConnection();

            /* ---------- Find keys to add ---------- */
            HashSet<String> schemasToCheck = new HashSet<String>(); // schema in which could be keys to add.
            ArrayList<Key> keysToCheck = new ArrayList<Key>();
            for (Table table : tablesToAdd) {
                String schemaName = table.getSchemaName();
                if (!schemasToCheck.contains(schemaName)) {
                    schemasToCheck.add(schemaName);
                    keysToCheck.addAll(tapSchema.getSchema(schemaName).getKeys());
                }
            }

            ArrayList<Key> keysToAdd = new ArrayList<Key>();
            for (Key key : keysToCheck) {
                for (Table table : tablesToAdd) {
                    String tableName = table.getFullName();
                    if (key.getFromTable().equals(tableName) || key.getTargetTable().equals(tableName)) {
                        keysToAdd.add(key);
                    }
                }
            }

            if (tapSchema.toCreate()) {
                createTapSchema(connection);
            }

            // Use tapSchema database
            connection.setCatalog(tapSchema.getName());

            /* ---------- Find keys to remove ---------- */
            if (tablesToRemove.size() > 0) {

                String tablesToRemoveStr = "";
                for (int i = 0; i < tablesToRemove.size(); i++) {
                    tablesToRemoveStr += "'" + tablesToRemove.get(i) + "'";
                    if (i < tablesToRemove.size() - 1) {
                        tablesToRemoveStr += ",";
                    }
                }

                (new ResultSetReader("SELECT CAST(key_id AS UNSIGNED) key_id FROM `keys`\n"
                        + "WHERE from_table IN (" + tablesToRemoveStr + ") OR target_table IN (" + tablesToRemoveStr + ")") {

                            @Override
                            public void manipulateItem(ResultSet resultSet) throws SQLException {
                                keysToRemove.add(resultSet.getInt("key_id"));
                            }
                        }).read(connection);

            }

            int newKeyId = 1; // key id is varchar and needs to be incremented manually. This is the start value.

            /* ---------- Find max key id after keys removal ---------- */
            // if some keys are already saved in the database we need to find their max value
            if (!tapSchema.toCreate()) {
                String query = "SELECT MAX(CAST(key_id AS UNSIGNED)) max_key from `keys`";

                if (keysToRemove.size() > 0) {
                    String keysToRemoveStr = "";
                    for (int i = 0; i < keysToRemove.size(); i++) {
                        keysToRemoveStr += "'" + keysToRemove.get(i) + "'";
                        if (i < keysToRemove.size() - 1) {
                            keysToRemoveStr += ",";
                        }
                    }

                    query += " WHERE key_id NOT IN (" + keysToRemoveStr + ")";
                }

                Statement s = null;
                ResultSet rs = null;
                try {
                    s = connection.createStatement();
                    rs = s.executeQuery(query);
                    while (rs.next()) {
                        newKeyId = rs.getInt("max_key") + 1;
                    }
                } catch (SQLException e) {
                    throw e;
                } finally {
                    if (rs != null) {
                        rs.close();
                    }
                    if (s != null) {
                        s.close();
                    }
                }
            }

            // Start update
            connection.setAutoCommit(false); // start transaction

            // REMOVE ELEMENTS
            if (!tapSchema.toCreate()) {
                for (int keyId : keysToRemove) {
                    statement = connection.prepareStatement("DELETE FROM key_columns WHERE key_id = ?");
                    statement.setString(1, keyId + "");
                    statement.executeUpdate();

                    statement = connection.prepareStatement("DELETE FROM `keys` WHERE key_id = ?");
                    statement.setString(1, keyId + "");
                    statement.executeUpdate();
                }

                for (String columnName : columnsToRemove) {
                    statement = connection.prepareStatement("DELETE FROM `columns` WHERE column_name = ?");
                    statement.setString(1, columnName);
                    statement.executeUpdate();
                }

                for (String tableFullName : tablesToRemove) {
                    statement = connection.prepareStatement("DELETE FROM `tables` WHERE table_name = ?");
                    statement.setString(1, tableFullName);
                    statement.executeUpdate();
                }

                for (String schemaName : schemasToRemove) {
                    statement = connection.prepareStatement("DELETE FROM `schemas` WHERE schema_name = ?");
                    statement.setString(1, schemaName);
                    statement.executeUpdate();
                }
            }

            // INSERT ELEMENTS
            for (Schema schema : schemasToAdd) {
                schema.insert(connection);
            }

            for (Table table : tablesToAdd) {
                table.insert(connection);
            }

            for (Column column : columnsToAdd) {
                column.insert(connection);
            }

            for (Key key : keysToAdd) {
                statement = connection.prepareStatement("INSERT INTO `keys` "
                        + "(key_id, from_table, target_table, utype, description, keyID) "
                        + "VALUES(?,?,?,null,null,null)");
                statement.setString(1, newKeyId + "");
                statement.setString(2, key.getFromTable());
                statement.setString(3, key.getTargetTable());
                statement.executeUpdate();
                statement.close();

                for (KeyColumn kc : key.getKeyColumns()) {
                    statement = connection.prepareStatement("INSERT INTO key_columns "
                            + "(key_id, from_column, target_column, key_columnID) "
                            + "VALUES(?,?,?,null)");
                    statement.setString(1, newKeyId + "");
                    statement.setString(2, kc.getFromColumn());
                    statement.setString(3, kc.getTargetColumn());
                    statement.executeUpdate();
                    statement.close();
                }

                newKeyId++;
            }

            //UPDATE ELEMENTS
            if (!tapSchema.toCreate()) {
                for (Schema schema : schemasToUpdate) {
                    schema.update(connection);
                }

                for (Table table : tablesToUpdate) {
                    table.update(connection);
                }

                for (Column column : columnsToUpdate) {
                    column.update(connection);
                }
            }

            connection.commit();
            
            tapSchema.update();
        } catch (SQLException e) {
            try {
                if (connection != null && !connection.getAutoCommit()) {
                    connection.rollback();
                }
            } catch (SQLException ex) {
                ex.printStackTrace(System.err);
            }
            e.printStackTrace(System.err);
        } finally {
            try {
                if (resultSet != null) {
                    resultSet.close();
                }
                if (statement != null) {
                    statement.close();
                }
                if (connection != null) {
                    connection.close();
                }
            } catch (SQLException ex) {
                ex.printStackTrace(System.err);
            }
        }
    }
}
